import logging

from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from rest_framework import status
from rest_framework.generics import get_object_or_404 as rest_get_object_or_404
from rest_framework.response import Response

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.storage.models import SharedUploadedFile
from mayan.apps.views.settings import setting_show_dropzone_submit_button

from ..classes import SourceBackendAction, SourceBackend
from ..forms import WebFormUploadFormHTML5
from ..tasks import task_process_document_upload

from .mixins import (
    SourceBackendCompressedMixin, SourceBackendInteractiveMixin,
    SourceBaseMixin
)

__all__ = ('SourceBackendWebForm',)
logger = logging.getLogger(name=__name__)


class SourceBackendWebForm(
    SourceBackendCompressedMixin, SourceBackendInteractiveMixin,
    SourceBaseMixin, SourceBackend
):
    actions = (
        SourceBackendAction(
            name='upload', arguments=(
                'document_type_id', 'expand'
            ), accept_files=True
        ),
    )
    label = _('Web form')
    upload_form_class = WebFormUploadFormHTML5

    def action_upload(self, document_type_id, file, request, expand=False):
        queryset = AccessControlList.objects.restrict_queryset(
            queryset=DocumentType.objects.all(),
            permission=permission_document_create,
            user=request.user
        )

        document_type = rest_get_object_or_404(
            pk=document_type_id, queryset=queryset
        )

        self.process_kwargs = {'request': request}

        shared_uploaded_file = SharedUploadedFile.objects.create(
            file=file
        )

        kwargs = {
            'callback_kwargs': self.get_callback_kwargs(),
            'document_type_id': document_type.pk,
            'expand': expand,
            'shared_uploaded_file_id': shared_uploaded_file.pk,
            'source_id': self.model_instance_id,
            'user_id': request.user.pk
        }

        task_process_document_upload.apply_async(kwargs=kwargs)

        return (
            None, Response(status=status.HTTP_202_ACCEPTED)
        )

    def get_shared_uploaded_files(self):
        return (
            SharedUploadedFile.objects.create(
                file=self.process_kwargs['forms']['source_form'].cleaned_data['file']
            ),
        )

    def get_view_context(self, context, request):
        if setting_show_dropzone_submit_button.value:
            form_disable_submit = False
        else:
            form_disable_submit = True

        return {
            'subtemplates_list': [
                {
                    'context': {
                        'forms': context['forms'],
                        'form_action': '{}?{}'.format(
                            reverse(
                                viewname=request.resolver_match.view_name,
                                kwargs=request.resolver_match.kwargs
                            ), request.META['QUERY_STRING']
                        ),
                        'form_css_classes': 'dropzone',
                        'form_disable_submit': form_disable_submit,
                        'form_id': 'html5upload',
                        'is_multipart': True
                    },
                    'name': 'appearance/generic_multiform_subtemplate.html'
                }
            ]
        }



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.queues import queue_documents

queue_documents.add_task_type(
    dotted_path='mayan.apps.document_downloads.tasks.task_document_file_compress',
    label=_('Generate a compressed document file bundle')
)



from django.contrib.contenttypes.models import ContentType

from mayan.apps.rest_api import generics

from .serializers import ContentTypeSerializer


class APIContentTypeList(generics.ListAPIView):
    """
    Returns a list of all the available content types.
    """
    serializer_class = ContentTypeSerializer
    source_queryset = ContentType.objects.order_by('app_label', 'model')



from rest_framework import status

from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..permissions import (
    permission_document_file_content_view,
    permission_document_type_parsing_setup
)

from .mixins import (
    DocumentFilePageContentAPITestMixin,
    DocumentTypeParsingSettingsAPIViewTestMixin
)


class DocumentFilePageContentAPITestCase(
    DocumentFilePageContentAPITestMixin, DocumentTestMixin, BaseAPITestCase
):
    def test_document_file_page_content_api_view_no_permission(self):
        self._clear_events()

        response = self._request_document_file_page_content_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_content_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_content_view
        )

        self._clear_events()

        response = self._request_document_file_page_content_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_page_content_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_content_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_document_file_page_content_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentTypeParsingSettingsAPIViewTestCase(
    DocumentTestMixin, DocumentTypeParsingSettingsAPIViewTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def test_document_type_parsing_settings_details_api_view_no_permission(self):
        self._clear_events()

        response = self._request_document_type_parsing_settings_details_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_parsing_settings_details_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_parsing_setup
        )

        self._clear_events()

        response = self._request_document_type_parsing_settings_details_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'auto_parsing': False})

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_parsing_settings_patch_api_view_no_permission(self):
        self._clear_events()

        response = self._request_document_type_parsing_settings_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_parsing_settings_patch_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_parsing_setup
        )

        self._clear_events()

        response = self._request_document_type_parsing_settings_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'auto_parsing': True})

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_parsing_settings_put_api_view_no_permission(self):
        self._clear_events()

        response = self._request_document_type_parsing_settings_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_parsing_settings_put_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_parsing_setup
        )

        self._clear_events()

        response = self._request_document_type_parsing_settings_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'auto_parsing': True})

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.db import models, transaction
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import (
    BackendModelMixin, ExtraDataModelMixin
)
from mayan.apps.events.classes import EventManagerSave
from mayan.apps.events.decorators import method_event

from .classes import SourceBackendNull
from .events import event_source_created, event_source_edited
from .managers import SourceManager
from .model_mixins import SourceBusinessLogicMixin


class Source(
    BackendModelMixin, ExtraDataModelMixin, SourceBusinessLogicMixin,
    models.Model
):
    _backend_model_null_backend = SourceBackendNull

    label = models.CharField(
        db_index=True, help_text=_('A short text to describe this source.'),
        max_length=128, unique=True, verbose_name=_('Label')
    )
    enabled = models.BooleanField(
        default=True, verbose_name=_('Enabled')
    )

    objects = SourceManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Source')
        verbose_name_plural = _('Sources')

    def __str__(self):
        return '%s' % self.label

    def delete(self, *args, **kwargs):
        with transaction.atomic():
            self.get_backend_instance().delete()
            super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_source_created,
            'target': 'self'
        },
        edited={
            'event': event_source_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        is_new = not self.pk

        with transaction.atomic():
            super().save(*args, **kwargs)

            self.get_backend_instance().clean()

            if is_new:
                self.get_backend_instance().create()
            else:
                self.get_backend_instance().save()



from django.db.models import Q

from ...models.workflow_state_escalation_models import WorkflowStateEscalation
from ...tasks import (
    task_workflow_instance_check_escalation,
    task_workflow_instance_check_escalation_all
)

from ..literals import (
    TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_AMOUNT,
    TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_AMOUNT_EDITED,
    TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_COMMENT,
    TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_UNIT
)


class WorkflowTemplateStateEscalationAPIViewTestMixin:
    def _request_test_workflow_template_state_escalation_create_api_view(self):
        data = {
            'amount': 1,
            'enabled': True,
            'workflow_template_transition_id': self._test_workflow_template_transition.pk
        }

        pk_list = list(
            WorkflowStateEscalation.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='rest_api:workflow-template-state-escalation-list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_state_id': self._test_workflow_template_states[0].pk
            }, data=data
        )

        try:
            self._test_workflow_estate_escalation = WorkflowStateEscalation.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowStateEscalation.DoesNotExist:
            self._test_workflow_estate_escalation = None

        return response

    def _request_test_workflow_template_state_escalation_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-state-escalation-detail', kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_state_id': self._test_workflow_template_states[0].pk,
                'workflow_template_state_escalation_id': self._test_workflow_template_state_escalation.pk
            }
        )

    def _request_test_workflow_template_state_escalation_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-state-escalation-detail', kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_state_id': self._test_workflow_template_states[0].pk,
                'workflow_template_state_escalation_id': self._test_workflow_template_state_escalation.pk
            }
        )

    def _request_test_workflow_template_state_escalation_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-state-escalation-detail', kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_state_id': self._test_workflow_template_states[0].pk,
                'workflow_template_state_escalation_id': self._test_workflow_template_state_escalation.pk
            }, data={
                'amount': TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_AMOUNT_EDITED
            }
        )

    def _request_test_workflow_template_state_escalation_edit_via_put_api_view(self):
        return self.put(
            viewname='rest_api:workflow-template-state-escalation-detail', kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_state_id': self._test_workflow_template_states[0].pk,
                'workflow_template_state_escalation_id': self._test_workflow_template_state_escalation.pk
            }, data={
                'amount': TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_AMOUNT_EDITED,
                'workflow_template_transition_id': self._test_workflow_template_transitions[0].pk
            }
        )

    def _request_test_workflow_template_state_escalation_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-state-escalation-list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_state_id': self._test_workflow_template_states[0].pk
            }
        )


class WorkflowTemplateStateEscalationTaskTestMixin:
    def _execute_task_workflow_instance_check_escalation(
        self, test_workflow_instance_id
    ):
        task_workflow_instance_check_escalation.apply_async(
            kwargs={
                'workflow_instance_id': test_workflow_instance_id
            }
        ).get()

    def _execute_task_workflow_instance_check_escalation_all(self):
        task_workflow_instance_check_escalation_all.apply_async().get()


class WorkflowTemplateStateEscalationTestMixin:
    def _create_test_workflow_template_state_escalation(self, extra_data=None):
        kwargs = {
            'amount': TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_AMOUNT,
            'comment': TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_COMMENT,
            'transition': self._test_workflow_template_transition,
            'unit': TEST_WORKFLOW_TEMPLATE_STATE_ESCALATION_UNIT
        }

        if extra_data:
            kwargs.update(extra_data)

        self._test_workflow_template_state_escalation = self._test_workflow_template_states[0].escalations.create(
            **kwargs
        )



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0077_favoritedocumentproxy'),
        ('checkouts', '0011_delete_newfileblock')
    ]

    operations = [
        migrations.AlterField(
            model_name='documentcheckout',
            name='document',
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='checkout', to='documents.document',
                verbose_name='Document'
            )
        )
    ]



from django.apps import apps
from django.utils.html import conditional_escape

from .events import event_parsing_document_file_submitted
from .tasks import task_parse_document_file


def method_document_content(self):
    file_latest = self.file_latest
    if file_latest:
        return file_latest.content()
    else:
        return []


def method_document_parsing_submit(self, user=None):
    latest_file = self.file_latest
    # Don't error out if document has no file.
    if latest_file:
        latest_file.submit_for_parsing(user=user)


def method_document_file_content(document_file):
    DocumentFilePageContent = apps.get_model(
        app_label='document_parsing', model_name='DocumentFilePageContent'
    )

    for page in document_file.pages.all():
        try:
            page_content = page.content.content
        except DocumentFilePageContent.DoesNotExist:
            """Page has no content, skip."""
        else:
            yield conditional_escape(text=str(page_content))


def method_document_file_parsing_submit(self, user=None):
    if user:
        user_id = user.pk
    else:
        user_id = None

    event_parsing_document_file_submitted.commit(
        action_object=self.document, actor=user, target=self
    )

    task_parse_document_file.apply_async(
        kwargs={'document_file_pk': self.pk, 'user_id': user_id}
    )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.search import (
    search_model_document, search_model_document_file,
    search_model_document_file_page, search_model_document_version,
    search_model_document_version_page
)
from mayan.apps.dynamic_search.search_models import SearchModel

from .permissions import permission_cabinet_view

# Cabinet

search_model_cabinet = SearchModel(
    app_label='cabinets', model_name='CabinetSearchResult',
    permission=permission_cabinet_view,
    serializer_path='mayan.apps.cabinets.serializers.CabinetSerializer'
)
search_model_cabinet.add_proxy_model(
    app_label='cabinets', model_name='Cabinet'
)

search_model_cabinet.add_model_field(field='label')

# Cabinet documents

search_model_cabinet.add_model_field(
    field='documents__document_type__label', label=_('Document type')
)
search_model_cabinet.add_model_field(
    field='documents__label', label=_('Document label')
)
search_model_cabinet.add_model_field(
    field='documents__description', label=_('Document description')
)
search_model_cabinet.add_model_field(
    field='documents__uuid', label=_('Document UUID')
)

# Cabinet documents files

search_model_cabinet.add_model_field(
    field='documents__files__checksum', label=_('Document file checksum')
)
search_model_cabinet.add_model_field(
    field='documents__files__mimetype', label=_('Document file MIME type')
)

# Document

search_model_document.add_model_field(
    field='cabinets__label', label=_('Cabinets')
)

# Document file

search_model_document_file.add_model_field(
    field='document__cabinets__label',
    label=_('Document cabinets')
)

# Document file page

search_model_document_file_page.add_model_field(
    field='document_file__document__cabinets__label',
    label=_('Document cabinets')
)

# Document version

search_model_document_version.add_model_field(
    field='document__cabinets__label',
    label=_('Document cabinets')
)

# Document version page

search_model_document_version_page.add_model_field(
    field='document_version__document__cabinets__label',
    label=_('Document cabinets')
)



import logging

from django.apps import apps

from mayan.apps.acls.models import AccessControlList
from mayan.apps.common.serialization import yaml_load

from ..classes import BaseDocumentFilenameGenerator
from ..permissions import permission_document_view
from ..settings import setting_language

logger = logging.getLogger(name=__name__)


class DocumentTypeBusinessLogicMixin:
    @property
    def trashed_documents(self):
        TrashedDocument = apps.get_model(
            app_label='documents', model_name='TrashedDocument'
        )

        return TrashedDocument.objects.filter(document_type=self)

    def get_document_count(self, user):
        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_document_view, queryset=self.documents,
            user=user
        )

        return queryset.count()

    def get_upload_filename(self, instance, filename):
        generator_klass = BaseDocumentFilenameGenerator.get(
            name=self.filename_generator_backend
        )
        generator_instance = generator_klass(
            **yaml_load(
                stream=self.filename_generator_backend_arguments or '{}'
            )
        )
        return generator_instance.upload_to(
            instance=instance, filename=filename
        )

    def new_document(
        self, file_object, label=None, description=None, language=None,
        user=None
    ):
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )

        try:
            document = Document(
                description=description or '', document_type=self,
                label=label or file_object.name,
                language=language or setting_language.value
            )
            document._event_keep_attributes = ('_event_actor',)
            document._event_actor = user
            document.save()
        except Exception as exception:
            logger.critical(
                'Unexpected exception while trying to create new document '
                '"%s" from document type "%s"; %s',
                label or file_object.name, self, exception
            )
            raise
        else:
            try:
                document_file = document.file_new(
                    file_object=file_object, filename=label, user=user
                )
            except Exception as exception:
                logger.critical(
                    'Unexpected exception while trying to create initial '
                    'file for document %s; %s',
                    label or file_object.name, exception
                )
                document.delete(to_trash=False)
                raise
            else:
                return (document, document_file)



from django import apps
from django.conf import settings
from django.test.runner import DiscoverRunner

from .literals import EXCLUDE_TEST_TAG


class NullMigrationsClass:
    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        return None


class MayanTestRunner(DiscoverRunner):
    @classmethod
    def add_arguments(cls, parser):
        DiscoverRunner.add_arguments(parser)
        parser.add_argument(
            '--mayan-apps', action='store_true', default=False,
            dest='mayan_apps',
            help='Test all Mayan apps that report to have tests.'
        )
        parser.add_argument(
            '--no-exclude', action='store_true', default=False,
            dest='no_exclude',
            help='Include excluded tests.'
        )
        parser.add_argument(
            '--skip-migrations', action='store_true', default=False,
            dest='skip_migrations', help='Skip execution of '
            'migrations after creating the database.'
        )

    def __init__(self, *args, **kwargs):
        self.mayan_apps = kwargs.pop('mayan_apps')
        self.no_exclude = kwargs.pop('no_exclude')
        if kwargs.pop('skip_migrations'):
            settings.MIGRATION_MODULES = NullMigrationsClass()

        super().__init__(*args, **kwargs)

        # Test that should be excluded by default
        # To include then pass --tag=exclude to the test runner invocation
        if not self.no_exclude:
            if EXCLUDE_TEST_TAG not in self.tags:
                self.exclude_tags |= {EXCLUDE_TEST_TAG}

    def build_suite(self, *args, **kwargs):
        # Apps that report they have tests
        if self.mayan_apps:
            args = list(args)
            args[0] = [
                app.name for app in apps.apps.get_app_configs() if getattr(
                    app, 'has_tests', False
                )
            ]

        return super().build_suite(*args, **kwargs)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('file_caching', '0008_auto_20210426_0717')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='cache',
            options={
                'ordering': ('id',), 'verbose_name': 'Cache',
                'verbose_name_plural': 'Caches'
            }
        )
    ]
